@@ -17,7 +17,14 @@
    Please see the accompanying COPYING file for the precise license terms. The
    GPL are also be read online at http://www.gnu.org/licenses/. */
 
+// Deal with Octave API breakage in different versions. At present, the
+// critical checkpoints are 3.8 and 4.2.
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=2
+#include "octave-config.h"
+#undef OCTAVE_USE_DEPRECATED_FUNCTIONS
+#else
 #include "octave/config.h"
+#endif
 
 #include <iostream>
 #include "octave.h"
@@ -26,7 +33,11 @@
 #if 0
 #include "unwind-prot.h"
 #endif
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=2
+#include "interpreter.h"
+#else
 #include "toplev.h"
+#endif
 #include "error.h" 
 #include "quit.h"
 #include "variables.h"
@@ -40,25 +51,97 @@
 
 #include "embed.h"
 
-/* Basic Octave interface. This is a heavily hacked version of octave_embed by
-   Paul Kienzle (http://wiki.octave.org/wiki.pl?OctaveEmbedded) which in turn
-   is based on Octave's toplevel. */
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=2
+// These aren't in the public API any more.
+extern "C" void octave_save_signal_mask (void);
+extern "C" void octave_restore_signal_mask (void);
+// These are named differently.
+#define can_interrupt octave::can_interrupt
+#define octave_catch_interrupts octave::catch_interrupts
+#define octave_interrupt_exception octave::interrupt_exception
+#define octave_execution_exception octave::execution_exception
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=3
+#define eval_string octave::eval_string
+#define feval octave::feval
+#endif
+#endif
+
+/* Basic Octave interface. This is a heavily hacked version of octave_embed
+   originally by Paul Kienzle (http://wiki.octave.org/wiki.pl?OctaveEmbedded)
+   which in turn is based on Octave's toplevel. */
 
 static bool init = false, first_init = false;
 
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=3
+// Octave 4.3+: Creating an embedded instance of the Octave interpreter is
+// child's play now. For now, we just emulate the old interface on top of
+// it. Also note that the command line arguments are always ignored now.
+
+#define octave_main my_octave_main
+#define clean_up_and_exit my_clean_up_and_exit
+#define octave_exit my_octave_exit
+
+static int octave_exit = 0;
+static octave::interpreter *embedded_interpreter = 0;
+
+static int my_octave_main(int argc, char **argv, int embedded)
+{
+  if (!embedded_interpreter) {
+    embedded_interpreter = new octave::interpreter();
+    embedded_interpreter->execute();
+  }
+}
+
+static void my_clean_up_and_exit(int exit_status, bool safe_to_return)
+{
+  if (embedded_interpreter) {
+    delete embedded_interpreter;
+    embedded_interpreter = 0;
+  }
+}
+
+// other recent breakage in the 4.3 API:
+#define is_bool_type islogical
+#define is_integer_type isinteger
+#define is_complex_type iscomplex
+
+#define get_global_value myget_global_value
+#define set_global_value myset_global_value
+
+octave_value
+get_global_value (const std::string& nm, bool silent)
+{
+  octave::symbol_table& symtab = embedded_interpreter->get_symbol_table();
+
+  octave_value val = symtab.global_varval (nm);
+
+  if (val.is_undefined () && ! silent)
+    error ("get_global_value: undefined symbol '%s'", nm.c_str ());
+
+  return val;
+}
+
+void
+set_global_value (const std::string& nm, const octave_value& val)
+{
+  octave::symbol_table& symtab = embedded_interpreter->get_symbol_table();
+
+  symtab.global_assign (nm, val);
+}
+#endif
+
 static void install_builtins();
 
 void octave_init(int argc, char *argv[])
 {
   if (!init) {
-#ifdef OCTAVE_3_8_PLUS
+    // We don't support a restart of the Octave interpreter, because that
+    // trips different kinds of issues with different versions of Octave
+    // (including segfaults in some versions). So just bail out here.
     if (first_init) {
-      // octave_main() segfaults when called a second time, so let's at least
-      // try to terminate gracefully here.
-      fprintf(stderr, "error: octave_init called twice, exiting\n");
-      exit(1);
+      fprintf(stderr, "error: octave_init called twice, ignored\n");
+      return;
     }
-#endif
     octave_main(argc,argv,1);
     init = true;
     if (!first_init) {
@@ -72,7 +155,7 @@ void octave_init(int argc, char *argv[])
 void octave_fini(void)
 {
   if (init) {
-#ifdef OCTAVE_3_8_PLUS
+#if OCTAVE_MAJOR>3 || OCTAVE_MAJOR>=3 && OCTAVE_MINOR>=8
     // Octave 3.8 doesn't expose do_octave_atexit() any more, so we call
     // clean_up_and_exit() instead, and prevent Octave from exiting the
     // process.
@@ -85,13 +168,18 @@ void octave_fini(void)
   }
 }
 
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=3
+#define recover octave::interpreter::recover_from_exception
+#elif OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=2
+#define recover recover_from_exception
+#else
 static void recover(void)
 {
 #if 0
   // This isn't supported in the latest Octave versions. We simply leave this
-  // disabled for now, which means that you'll have to use 'unwind_protect'
-  // explicitly in your Octave code in order to handle Octave exceptions.
-  // XXXFIXME: This might leak memory in some cases??
+  // disabled, which means that you'll have to use 'unwind_protect' explicitly
+  // in your Octave code in order to handle Octave exceptions. XXXFIXME: This
+  // might leak memory in some cases??
   unwind_protect::run_all ();
 #endif
   can_interrupt = true;
@@ -102,76 +190,72 @@ static void recover(void)
   octave_restore_signal_mask ();
   octave_catch_interrupts ();
 }
+#endif
 
 int octave_eval(const char *cmd)
 {
-  int parse_status;
+  int parse_status = 0;
 
+  if (!init) return -1;
+
+  // XXXFIXME: Need to look at this because octave_save_signal_mask() and
+  // octave_restore_signal_mask() are not in the public API of Octave 4.2+ any
+  // more. They're still in the library, though, and we (presumably) need them
+  // here so that we can set up an interrupt (SIG_INT) handler.
   octave_save_signal_mask ();
-  if (octave_set_current_context)
-    {
+
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=2
+  // Get rid of all the cruft below which is in Kienzle's original code but
+  // doesn't seem to be needed any more.
+#else
+  // XXXFIXME: Really need to clean up this cruft, this just seems needlessly
+  // complicated. Is this even needed any more?
+  if (octave_set_current_context) {
 #if defined (USE_EXCEPTIONS_FOR_INTERRUPTS)
-      panic_impossible ();
+    panic_impossible ();
 #else
 #if 0
-      unwind_protect::run_all ();
+    unwind_protect::run_all ();
 #endif
-      raw_mode (0);
-      std::cout << "\n";
-      octave_restore_signal_mask ();
+    raw_mode (0);
+    std::cout << "\n";
+    octave_restore_signal_mask ();
+#endif
+  }
 #endif
-    }
 
   can_interrupt = true;
   octave_catch_interrupts ();
   octave_initialized = true;
 
-  // XXX FIXME XXX need to give caller an opaque pointer
-  // so that they can define and use separate namespaces
-  // when calling octave, with a shared global namespace.
-  // Something like:
-  //   int call_octave (const char *string, void *psymtab = NULL) {
-  //     ...
-  //     curr_sym_tab = psymtab == NULL ? top_level_sym_tab : symbol_table;
-  // I suppose to be safe from callbacks (surely we have to
-  // provide some way to call back from embedded octave into
-  // the user's application), we should push and pop the current
-  // symbol table.
-
-  // Note that I'm trying to distinguish exception from 
-  // failure in the return codes. I believe failure is 
-  // indicated by -1.  I have execution exception (including
-  // user interrupt and more dramatic failures) returning -2
-  // and memory failure returning -3.  We should formalize
-  // this with error codes defined in embed_octave.h.  Maybe
-  // a more fine-grained approach could be used within octave
-  // proper.
-  try 
-    {
-#if 0
-      curr_sym_tab = top_level_sym_tab;
+  // Note that Kienzle's code uses various error codes here: -1 indicates
+  // "failure", which seems to be set in the interpreter somewhere, -2 is user
+  // interrupt (and more dramatic failures), -3 memory failure.
+  try {
+    reset_error_handler ();
+    // XXXFIXME: eval_string() just always segfaults with Octave 4.2.0+. No
+    // workaround for this is known yet. Octave 4.0 and 4.3 seem to be ok,
+    // though, so just stick to one of these until this is fixed.
+    eval_string(cmd, false, parse_status, 0);
+  } catch (octave_interrupt_exception) {
+    recover ();
+    std::cout << "\n";
+    error_state = -2;
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=2
+  } catch (octave_execution_exception) {
+    recover ();
+    std::cout << "\n";
+    error_state = -1;
 #endif
-      reset_error_handler ();
-      eval_string(cmd, false, parse_status, 0);
-    }
-  catch (octave_interrupt_exception)
-    {
-      recover ();
-      std::cout << "\n"; 
-      error_state = -2; 
-    }
-  catch (std::bad_alloc)
-    {
-      recover ();
-      std::cout << "\n"; 
-      error_state = -3;
-    }
+  } catch (std::bad_alloc) {
+    recover ();
+    std::cout << "\n";
+    error_state = -3;
+  }
 
-  octave_restore_signal_mask();
+  octave_restore_signal_mask ();
   octave_initialized = false;
 
-  // XXX FIXME XXX callbacks calling embed_octave
-  // may or may not want error_state reset.
   return error_state;
 }
 
@@ -220,17 +304,6 @@ gsl_matrix_calloc(const size_t n1, const size_t n2)
   return m;
 }
 
-#if 0
-static void gsl_matrix_free(gsl_matrix *m)
-{
-  if (m->owner) {
-    if (m->block) free(m->block->data);
-    free(m->block);
-  }
-  free(m);
-}
-#endif
-
 static gsl_matrix_complex* 
 gsl_matrix_complex_alloc(const size_t n1, const size_t n2)
 {
@@ -271,17 +344,6 @@ gsl_matrix_complex_calloc(const size_t n1, const size_t n2)
   return m;
 }
 
-#if 0
-static void gsl_matrix_complex_free(gsl_matrix_complex *m)
-{
-  if (m->owner) {
-    if (m->block) free(m->block->data);
-    free(m->block);
-  }
-  free(m);
-}
-#endif
-
 static gsl_matrix_int* 
 gsl_matrix_int_alloc(const size_t n1, const size_t n2)
 {
@@ -322,17 +384,6 @@ gsl_matrix_int_calloc(const size_t n1, const size_t n2)
   return m;
 }
 
-#if 0
-static void gsl_matrix_int_free(gsl_matrix_int *m)
-{
-  if (m->owner) {
-    if (m->block) free(m->block->data);
-    free(m->block);
-  }
-  free(m);
-}
-#endif
-
 /* GSL doesn't really support empty matrices, so we fake them by allocating 1
    dummy row or column if the corresponding dimension is actually zero. */
 static inline gsl_matrix*
@@ -809,12 +860,14 @@ static octave_value *pure_to_octave(pure_expr *x)
 
 pure_expr *octave_get(const char *id)
 {
+  if (!init) return 0;
   octave_value val = get_global_value(id, true);
   return octave_to_pure(val);
 }
 
 pure_expr *octave_set(const char *id, pure_expr *x)
 {
+  if (!init) return 0;
   octave_value *val = pure_to_octave(x);
   if (val) {
     set_global_value(id, *val);
@@ -833,10 +886,16 @@ pure_expr *octave_set(const char *id, pure_expr *x)
 
 pure_expr *octave_func(pure_expr *fun)
 {
+  if (!init) return 0;
   // First try to find an ordinary function handle.
   char *s;
   if (pure_is_cstring_dup(fun, &s)) {
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=3
+    octave::symbol_table& symtab = embedded_interpreter->get_symbol_table();
+    octave_value f = symtab.find_function(s);
+#else
     octave_value f = symbol_table::find_function(s);
+#endif
     free(s);
     if (f.is_defined()) {
       pure_expr *f = pure_string_dup("str2func"),
@@ -863,6 +922,7 @@ pure_expr *octave_func(pure_expr *fun)
 
 pure_expr *octave_call(pure_expr *fun, int nargout, pure_expr *args)
 {
+  if (!init) return 0;
   pure_expr **xs;
   size_t n;
   octave_value *v;
@@ -998,7 +1058,13 @@ DEFUN_DLD(pure_call, args, nargout, PURE_HELP)
 
 static void install_builtins()
 {
-#ifdef OCTAVE_3_8_PLUS
+#if OCTAVE_MAJOR>4 || OCTAVE_MAJOR>=4 && OCTAVE_MINOR>=3
+  // install_builtin_function is deprecated in 4.3, might be gone in 4.4+, but
+  // this will hopefully continue to work
+  octave_value fcn (new octave_builtin (Fpure_call, "pure_call", "embed.cc", PURE_HELP));
+  octave::symbol_table& symtab = embedded_interpreter->get_symbol_table();
+  symtab.install_built_in_function ("pure_call", fcn);
+#elif OCTAVE_MAJOR>3 || OCTAVE_MAJOR>=3 && OCTAVE_MINOR>=8
   install_builtin_function(Fpure_call, "pure_call", "embed.cc", PURE_HELP);
 #else
   install_builtin_function(Fpure_call, "pure_call", PURE_HELP);
